@@ -16,6 +16,7 @@
 #include "backwards_compat.c"
 
 #define streq(a,b) (strcmp((a),(b)) == 0)
+#define strstarts(a,start) (strncmp((a),(start), strlen(start)) == 0)
 
 #ifndef MODULE_DIR
 #define MODULE_DIR "/lib/modules"
@@ -115,6 +116,33 @@ static const char *next_string(const char *string, unsigned long *secsize)
 	return string;
 }
 
+struct param
+{
+	struct param *next;
+	const char *name; 	/* Terminated by a colon */
+	const char *param;
+	const char *type;
+};
+
+static struct param *add_param(const char *name, struct param **list)
+{
+	struct param *i;
+	unsigned int namelen = strcspn(name, ":") + 1;
+
+	for (i = *list; i; i = i->next)
+		if (strncmp(i->name, name, namelen) == 0)
+			return i;
+	i = malloc(sizeof(*i) + namelen+1);
+	strncpy((char *)(i + 1), name, namelen);
+	((char *)(i + 1))[namelen] = '\0';
+	i->name = (char *)(i + 1);
+	i->param = NULL;
+	i->type = NULL;
+	i->next = *list;
+	*list = i;
+	return i;
+}
+
 static void print_tag(const char *tag, const char *info, unsigned long size,
 		      const char *filename, char sep)
 {
@@ -133,9 +161,27 @@ static void print_tag(const char *tag, const char *info, unsigned long size,
 static void print_all(const char *info, unsigned long size,
 		      const char *filename, char sep)
 {
+	struct param *i, *params = NULL;
+
 	printf("%-16s%s%c", "filename:", filename, sep);
 	for (; info; info = next_string(info, &size)) {
-		char *eq;
+		char *eq, *colon;
+
+		/* We expect this in parm and parmtype. */
+		colon = strchr(info, ':');
+
+		/* We store these for handling at the end */
+		if (strstarts(info, "parm=") && colon) {
+			i = add_param(info + strlen("parm="), &params);
+			i->param = colon + 1;
+			continue;
+		}
+		if (strstarts(info, "parmtype=") && colon) {
+			i = add_param(info + strlen("parmtype="), &params);
+			i->type = colon + 1;
+			continue;
+		}
+
 		if (!sep) {
 			printf("%s%c", info, sep);
 			continue;
@@ -151,6 +197,17 @@ static void print_all(const char *info, unsigned long size,
 			printf("%-16s%s%c", tag, eq+1, sep);
 		}
 	}
+
+	/* Now show parameters. */
+	for (i = params; i; i = i->next) {
+		if (!i->param)
+			printf("%-16s%s%s%c", "parm:", i->name, i->type, sep);
+		else if (i->type)
+			printf("%-16s%s%s (%s)%c",
+			       "parm:", i->name, i->param, i->type, sep);
+		else 
+			printf("%-16s%s%s%c", "parm:", i->name, i->param, sep);
+	}
 }
 
 static struct option options[] =
@@ -215,7 +272,8 @@ static char *next_line(char *p, const char *end)
 	return (char *)end + 1;
 }
 
-static void *grab_module(const char *name, unsigned long *size, char**filename)
+static void *grab_module(const char *name, unsigned long *size, char**filename,
+	const char *kernel)
 {
 	char *data;
 	struct utsname buf;
@@ -233,8 +291,13 @@ static void *grab_module(const char *name, unsigned long *size, char**filename)
 	}
 
 	/* Search for it in modules.dep. */
-	uname(&buf);
-	asprintf(&depname, "%s/%s/modules.dep", MODULE_DIR, buf.release);
+	if (kernel) {
+		asprintf(&depname, "%s/%s/modules.dep", MODULE_DIR, kernel);
+	} else {
+		uname(&buf);
+		asprintf(&depname, "%s/%s/modules.dep", MODULE_DIR,
+			 buf.release);
+	}
 	data = grab_file(depname, size);
 	if (!data) {
 		fprintf(stderr, "modinfo: could not open %s\n", depname);
@@ -265,7 +328,7 @@ static void *grab_module(const char *name, unsigned long *size, char**filename)
 
 static void usage(const char *name)
 {
-	fprintf(stderr, "Usage: %s [-0][-F field] module...\n"
+	fprintf(stderr, "Usage: %s [-0][-F field][-k kernelversion] module...\n"
 		" Prints out the information about one or more module(s).\n"
 		" If a fieldname is given, just print out that field (or nothing if not found).\n"
 		" Otherwise, print all information out in a readable form\n"
@@ -277,6 +340,7 @@ int main(int argc, char *argv[])
 {
 	union { short s; char c[2]; } endian_test;
 	const char *field = NULL;
+	const char *kernel = NULL;
 	char sep = '\n';
 	unsigned long infosize;
 	int opt, ret = 0;
@@ -290,7 +354,7 @@ int main(int argc, char *argv[])
 	else
 		abort();
 
-	while ((opt = getopt_long(argc,argv,"adlpVhn0F:",options,NULL)) >= 0){
+	while ((opt = getopt_long(argc,argv,"adlpVhn0F:k:",options,NULL)) >= 0){
 		switch (opt) {
 		case 'a': field = "author"; break;
 		case 'd': field = "description"; break;
@@ -300,6 +364,7 @@ int main(int argc, char *argv[])
 		case 'V': printf(PACKAGE " version " VERSION "\n"); exit(0);
 		case 'F': field = optarg; break;
 		case '0': sep = '\0'; break;
+		case 'k': kernel = optarg; break;
 		default:
 			usage(argv[0]); exit(0);
 		}
@@ -312,19 +377,23 @@ int main(int argc, char *argv[])
 		unsigned long modulesize;
 		char *filename;
 
-		mod = grab_module(argv[opt], &modulesize, &filename);
+		mod = grab_module(argv[opt], &modulesize, &filename, kernel);
 		if (!mod) {
 			ret = 1;
 			continue;
 		}
 
 		info = get_section(mod, modulesize, &infosize, ".modinfo");
-		if (!info)
+		if (!info) {
+			release_file(mod, modulesize);
+			free(filename);
 			continue;
+		}
 		if (field)
 			print_tag(field, info, infosize, filename, sep);
 		else
 			print_all(info, infosize, filename, sep);
+		release_file(mod, modulesize);
 		free(filename);
 	}
 	return ret;
